@@ -26,7 +26,7 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 from __future__ import unicode_literals
 
-__version__ = "0.4.2"
+__version__ = "0.4.3"
 __author__ = "np1"
 __license__ = "LGPLv3"
 
@@ -93,7 +93,7 @@ def fetch_decode(url, encoding=None):
             time.sleep(.5)
             return fetch_decode(url, encoding)
         else:
-            raise e
+            raise
 
     ct = req.headers['content-type']
 
@@ -110,6 +110,32 @@ def fetch_decode(url, encoding=None):
         return req.read()
 
 
+class GdataError(Exception):
+    """Gdata query failed."""
+    pass
+
+
+def call_gdata(api, qs):
+    """Make a request to the youtube gdata api."""
+    qs = dict(qs)
+    qs['key'] = g.api_key
+    url = g.urls['gdata'] + api + '?' + urlencode(qs)
+
+    try:
+        data = g.opener.open(url).read().decode()
+
+    except HTTPError as e:
+        try:
+            errdata = e.file.read().decode()
+            error = json.loads(errdata)['error']['message']
+            errmsg = 'Youtube Error %d: %s' % (e.getcode(), error)
+        except:
+            errmsg = str(e)
+        raise GdataError(errmsg)
+
+    return json.loads(data)
+
+
 def new(url, basic=True, gdata=False, signature=True, size=False,
         callback=None, ydl_opts=None):
     """ Return a new pafy instance given a url or video id.
@@ -148,10 +174,8 @@ def get_video_gdata(video_id):
     new.callback("Fetching video gdata")
     query = {'part': 'id,snippet,statistics',
              'maxResults': 1,
-             'id': video_id,
-             'key': g.api_key}
-    url = g.urls['gdata'] + '?' + urlencode(query)
-    gdata = fetch_decode(url)  # unicode
+             'id': video_id}
+    gdata = call_gdata('videos', query)
     dbg("Fetched video gdata")
     new.callback("Fetched video gdata")
     return gdata
@@ -186,9 +210,8 @@ class g(object):
     """ Class for holding constants needed throughout the module. """
 
     urls = {
-        'gdata': "https://www.googleapis.com/youtube/v3/videos",
+        'gdata': "https://www.googleapis.com/youtube/v3/",
         'watchv': "http://www.youtube.com/watch?v=%s",
-        'vidcat': "https://www.googleapis.com/youtube/v3/videoCategories",
         'playlist': ('http://www.youtube.com/list_ajax?'
                      'style=json&action_get_list=1&list=%s'),
         'thumb': "http://i.ytimg.com/vi/%s/default.jpg",
@@ -485,7 +508,7 @@ class Stream(object):
             elapsed = time.time() - t0
             bytesdone += len(chunk)
             if elapsed:
-                rate = ((bytesdone - offset) / 1024) / elapsed
+                rate = ((float(bytesdone) - float(offset)) / 1024.0) / elapsed
                 eta = (total - bytesdone) / (rate * 1024)
             else: # Avoid ZeroDivisionError
                 rate = 0
@@ -562,7 +585,6 @@ class Pafy(object):
         self._bigthumbhd = None
         self._mix_pl = None
         self.expiry = None
-        self.playlist_meta = None
 
         self._ydl_opts = g.def_ydl_opts
         if ydl_opts:
@@ -614,8 +636,7 @@ class Pafy(object):
         if self._have_gdata:
             return
 
-        gdata = get_video_gdata(self.videoid)
-        item = json.loads(gdata)['items'][0]
+        item = get_video_gdata(self.videoid)['items'][0]
         snippet = item['snippet']
         self._published = uni(snippet['publishedAt'])
         self._description = uni(snippet["description"])
@@ -916,9 +937,7 @@ class Pafy(object):
         self._rating = pl_data.get("rating", 0.0)
         self._viewcount = "".join(re.findall(r"\d", pl_data.get("views", "0")))
         self._viewcount = int(self._viewcount)
-        self._thumb = pl_data.get("thumbnail")
         self._description = pl_data.get("description")
-        self.playlist_meta = pl_data
 
 
 def get_categoryname(cat_id):
@@ -929,12 +948,9 @@ def get_categoryname(cat_id):
     if cached.get('updated', 0) > timestamp - g.lifespan:
         return cached.get('title', 'unknown')
     # call videoCategories API endpoint to retrieve title
-    url = g.urls['vidcat']
     query = {'id': cat_id,
-             'part': 'snippet',
-             'key': g.api_key}
-    url += "?" + urlencode(query)
-    catinfo = json.loads(fetch_decode(url))
+             'part': 'snippet'}
+    catinfo = call_gdata('videoCategories', query)
     try:
         for item in catinfo.get('items', []):
             title = item.get('snippet', {}).get('title', 'unknown')
@@ -956,12 +972,9 @@ def set_categories(categories):
     idlist = [cid for cid, item in categories.items()
               if item.get('updated', 0) < timestamp - g.lifespan]
     if len(idlist) > 0:
-        url = g.urls['vidcat']
         query = {'id': ','.join(idlist),
-                 'part': 'snippet',
-                 'key': g.api_key}
-        url += "?" + urlencode(query)
-        catinfo = json.loads(fetch_decode(url))
+                 'part': 'snippet'}
+        catinfo = call_gdata('videoCategories', query)
         try:
             for item in catinfo.get('items', []):
                 cid = item['id']
@@ -982,17 +995,7 @@ def dump_cache():
     return g.cache
 
 
-def get_playlist(playlist_url, basic=False, gdata=False, signature=True,
-                 size=False, callback=lambda x: None):
-    """ Return a dict containing Pafy objects from a YouTube Playlist.
-
-    The returned Pafy objects are initialised using the arguments to
-    get_playlist() in the manner documented for pafy.new()
-
-    """
-    # pylint: disable=R0914
-    # too many local vars
-
+def extract_playlist_id(playlist_url):
     # Normal playlists start with PL, Mixes start with RD + first video ID
     idregx = re.compile(r'((?:RD|PL)[-_0-9a-zA-Z]+)$')
 
@@ -1008,6 +1011,20 @@ def get_playlist(playlist_url, basic=False, gdata=False, signature=True,
         if 'list' in query and idregx.match(query['list'][0]):
             playlist_id = query['list'][0]
 
+    return playlist_id
+
+
+def get_playlist(playlist_url, basic=False, gdata=False, signature=True,
+                 size=False, callback=lambda x: None):
+    """ Return a dict containing Pafy objects from a YouTube Playlist.
+
+    The returned Pafy objects are initialised using the arguments to
+    get_playlist() in the manner documented for pafy.new()
+
+    """
+
+    playlist_id = extract_playlist_id(playlist_url)
+
     if not playlist_id:
         err = "Unrecognized playlist url: %s"
         raise ValueError(err % playlist_url)
@@ -1078,6 +1095,125 @@ def get_playlist(playlist_url, basic=False, gdata=False, signature=True,
     return playlist
 
 
+def parseISO8591(duration):
+    """ Parse ISO 8591 formated duration """
+    regex = re.compile(r'PT((\d{1,3})H)?((\d{1,3})M)?((\d{1,2})S)?')
+    if duration:
+        duration = regex.findall(duration)
+        if len(duration) > 0:
+            _, hours, _, minutes, _, seconds = duration[0]
+            duration = [seconds, minutes, hours]
+            duration = [int(v) if len(v) > 0 else 0 for v in duration]
+            duration = sum([60**p*v for p, v in enumerate(duration)])
+        else:
+            duration = 30
+    else:
+        duration = 30
+    return duration
+
+
+class Playlist(object):
+    _items = None
+
+    def __init__(self, playlist_url, basic, gdata, signature, size, callback):
+        playlist_id = extract_playlist_id(playlist_url)
+
+        if not playlist_id:
+            err = "Unrecognized playlist url: %s"
+            raise ValueError(err % playlist_url)
+
+        query = {'part': 'snippet, contentDetails',
+                'id': playlist_id}
+        allinfo = call_gdata('playlists', query)
+
+        pl = allinfo['items'][0]
+
+        self.plid = playlist_id
+        self.title = pl['snippet']['title']
+        self.author = pl['snippet']['channelTitle']
+        self.description = pl['snippet']['description']
+        self ._len = pl['contentDetails']['itemCount']
+        self._basic = basic
+        self._gdata = gdata
+        self._signature = signature
+        self._size = size
+        self._callback = callback
+
+    def __len__(self):
+        return self._len
+    
+    def __iter__(self):
+        if self._items is not None:
+            for  i in self._items:
+                yield i
+            return
+
+        items = []
+
+        # playlist items specific metadata
+        query = {'part': 'snippet',
+                'maxResults': 50,
+                'playlistId': self.plid}
+
+        while True:
+            playlistitems = call_gdata('playlistItems', query)
+
+            query2 = {'part':'contentDetails,snippet,statistics',
+                      'maxResults': 50,
+                      'id': ','.join(i['snippet']['resourceId']['videoId']
+                          for i in playlistitems['items'])}
+            wdata = call_gdata('videos', query2)
+
+            for v, vextra in zip(playlistitems['items'], wdata['items']):
+                vid_data = dict(
+                    title=v['snippet']['title'],
+                    author=v['snippet']['channelTitle'],
+                    thumbnail=v['snippet'].get('thumbnails', {}
+                        ).get('default', {}).get('url'),
+                    description=v['snippet']['description'],
+                    length_seconds=parseISO8591(
+                        vextra['contentDetails']['duration']),
+                    category=get_categoryname(vextra['snippet']['categoryId']),
+                    views=vextra['statistics'].get('viewCount',0),
+                    likes=vextra['statistics'].get('likeCount',0),
+                    dislikes=vextra['statistics'].get('dislikeCount',0),
+                    comments=vextra['statistics'].get('commentCount',0),
+                )
+
+                try:
+                    pafy_obj = new(v['snippet']['resourceId']['videoId'],
+                            basic=self._basic, gdata=self._gdata,
+                            signature=self._signature, size=self._size,
+                            callback=self._callback)
+
+                except IOError as e:
+                    self.callback("%s: %s" % (v['title'], e.message))
+                    continue
+
+                pafy_obj.populate_from_playlist(vid_data)
+                items.append(pafy_obj)
+                self._callback("Added video: %s" % vid_data['title'])
+                yield pafy_obj
+
+            if not playlistitems.get('nextPageToken'):
+                break
+            query['pageToken'] = playlistitems['nextPageToken']
+
+        self._items = items
+
+
+def get_playlist2(playlist_url, basic=False, gdata=False, signature=True,
+                 size=False, callback=lambda x: None):
+    """ Return a Playlist object from a YouTube Playlist.
+
+    The returned Pafy objects are initialised using the arguments to
+    get_playlist() in the manner documented for pafy.new()
+
+    """
+
+    return Playlist(playlist_url, basic, gdata, signature, size, callback)
+
+
 def set_api_key(key):
     """Sets the api key to be used with youtube."""
     g.api_key = key
